"""Functions for measuring column width of characters."""

__all__ = ["char_width", "str_width"]

def char_width(char: str) -> int:
    """
    Return the column width of a character.

    If the length of ``char`` is greater than 1, only the column width of the first
    character is returned. If the length of ``char`` is 0, ``0`` is returned.

    Parameters
    ----------
    char : str
        A single character.

    Returns
    -------
    int
        The character column width.
    """

def str_width(chars: str) -> int:
    """
    Return the total column width of a string.

    Parameters
    ----------
    chars : str
        A string.

    Returns
    -------
    int
        The total column width of the string.
    """



"""
A growable string buffer.

This wrapper around ``fbuf`` should be removed shortly after sixel branch is merged and
Vt100Terminal can be rewritten in cython to handle `fbuf` directly.
"""

class FBufWrapper:
    """A growable string buffer."""

    def __len__(self) -> int:
        """Length of string buffer."""

    def __bool__(self) -> bool:
        """Whether string buffer is nonempty."""

    def write(self, s: bytes) -> None:
        """Write bytes to string buffer."""

    def flush(self) -> None:
        """Flush bytes to stdout."""



"""Platform specific VT100 terminals."""

import asyncio
import platform
import sys
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Final

from ..geometry import Size
from .events import DeviceAttributesReportEvent, Event, PixelGeometryReportEvent
from .vt100_terminal import DRS_REQUEST_TIMEOUT, Vt100Terminal

__all__ = [
    "Vt100Terminal",
    "app_mode",
    "get_platform_terminal",
    "get_sixel_info",
]

_SIXEL_SUPPORT: Final = 4
"""Terminal attribute for sixel support."""


@contextmanager
def app_mode(
    terminal: Vt100Terminal, event_handler: Callable[[list[Event]], None]
) -> Iterator[None]:
    """
    Put terminal into app mode and dispatch input events.

    Parameters
    ----------
    terminal : Vt100Terminal
        Terminal to put in app mode.
    event_handler : Callable[[list[Event]], None]
        A callable that handles terminal input events.
    """
    try:
        terminal.raw_mode()
        terminal.attach(event_handler)
        terminal.enable_mouse_support()
        terminal.enable_bracketed_paste()
        terminal.enable_reporting_focus()
        terminal.hide_cursor()
        terminal.flush()
        yield
    finally:
        if terminal.in_alternate_screen:
            terminal.exit_alternate_screen()
        else:
            terminal.erase_in_display()
        terminal.reset_attributes()
        terminal.show_cursor()
        terminal.disable_reporting_focus()
        terminal.disable_bracketed_paste()
        terminal.disable_mouse_support()
        terminal.flush()
        terminal.unattach()
        terminal.restore_console()


def get_platform_terminal() -> Vt100Terminal:
    """
    Return a platform-specific terminal.

    Returns
    -------
    Vt100Terminal
        A platform-specific VT100 terminal.

    Raises
    ------
    RuntimeError
        If terminal isn't interactive or terminal doesn't support VT100 sequences.
    """
    if not sys.stdin.isatty():
        raise RuntimeError("Terminal is non-interactive.")

    if platform.system() == "Windows":
        from .windows_terminal import WindowsTerminal, is_vt100_enabled

        if not is_vt100_enabled():
            raise RuntimeError("Terminal doesn't support VT100 sequences.")

        return WindowsTerminal()
    else:
        from .linux_terminal import LinuxTerminal

        return LinuxTerminal()


async def get_sixel_info(terminal: Vt100Terminal) -> tuple[bool, Size]:
    """
    Determine if terminal has sixel support and, if supported, its pixel geometry.

    Parameters
    ----------
    terminal : VT100Terminal
        A VT100 terminal to query for sixel support.

    Returns
    -------
    tuple[bool, Size]
        Whether sixel is supported and the pixel geometry.
    """
    sixel_support: bool = False
    pixel_geometry: Size = Size(20, 10)
    report_timeout: asyncio.TimerHandle
    terminal_info_reported: asyncio.Event = asyncio.Event()
    cell_reported: bool = False

    def report_handler(events: list[Event]) -> None:
        """Handle terminal reports."""
        nonlocal sixel_support, pixel_geometry, cell_reported, report_timeout

        for event in events:
            if isinstance(event, DeviceAttributesReportEvent):
                report_timeout.cancel()
                sixel_support = _SIXEL_SUPPORT in event.device_attributes
                if sixel_support:
                    terminal.request_pixel_geometry()
                    terminal.request_terminal_geometry()
                    report_timeout = asyncio.get_running_loop().call_later(
                        DRS_REQUEST_TIMEOUT, terminal_info_reported.set
                    )
                else:
                    terminal_info_reported.set()
            elif isinstance(event, PixelGeometryReportEvent):
                report_timeout.cancel()
                if event.kind == "cell":
                    pixel_geometry = event.geometry
                    cell_reported = True
                elif not cell_reported:
                    ph, pw = event.geometry
                    th, tw = terminal.get_size()
                    pixel_geometry = Size(ph // th, pw // tw)
                terminal_info_reported.set()

    old_handler = terminal._event_handler
    terminal._event_handler = report_handler
    terminal.request_device_attributes()
    report_timeout = asyncio.get_running_loop().call_later(
        DRS_REQUEST_TIMEOUT, terminal_info_reported.set
    )
    await terminal_info_reported.wait()
    terminal._event_handler = old_handler
    return sixel_support, pixel_geometry



"""Build cython extensions and generated files."""

import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import setup
from setuptools.command.build_py import build_py
from wcwidth import wcwidth

_CWIDTH_HEADER = """\
#ifndef _WIN32
    #include <stddef.h>
#endif
#include "cwidth.h"


int cwidth(uint32_t wc) {
    // Character widths as generated by wcwidth.
    // Each item in `CHAR_WIDTHS` represents an interval of ords with common width,
    // i.e., `{0u, 31u, 0u}` means ords 0 through 31 (inclusive) have width 0.
    // Intervals with width 1 are omitted so if an ord doesn't belong to any interval
    // we can assume it has width 1.
    static const uint32_t CHAR_WIDTHS[][3] = {
"""
_CWIDTH_FOOTER = """\
    };
    static const size_t CHAR_WIDTHS_LEN = sizeof(CHAR_WIDTHS) / sizeof(int[3]);

    size_t lo = 0, hi = CHAR_WIDTHS_LEN, mid;
    while(lo < hi) {
        mid = (lo + hi) / 2;
        if(wc < CHAR_WIDTHS[mid][0]) hi = mid;
        else if(wc > CHAR_WIDTHS[mid][1]) lo = mid + 1;
        else return CHAR_WIDTHS[mid][2];
    }
    return 1;
}
"""


def _create_cwidth():
    """
    Build ``cwidth.c``.

    This function builds the table used in ``cwidth.c`` needed to determine displayed
    width of a character in the terminal. The widths are taken directly from `wcwidth`,
    but the ``cwidth`` function is a couple of magnitude times faster than
    ``wcwidth.wcwidth``.
    """
    groups = []
    start = 0
    group_width = 1
    for codepoint in range(sys.maxunicode + 1):
        char_width = max(wcwidth(chr(codepoint)), 0)
        if char_width != group_width:
            if group_width != 1:
                groups.append((start, codepoint - 1, group_width))
            group_width = char_width
            start = codepoint

    if group_width != 1:
        groups.append((start, codepoint, group_width))

    with open("src/batgrl/cwidth.c", "w") as file:
        file.write(_CWIDTH_HEADER)
        for group in groups:
            file.write("        {{{}u, {}u, {}u}},\n".format(*group))
        file.write(_CWIDTH_FOOTER)


class build_py_with_cwidth(build_py):
    """Generate ``cwidth.c`` on build."""

    def run(self):
        """Generate ``cwidth.c`` on build."""
        super().run()
        _create_cwidth()


setup(
    ext_modules=cythonize(
        [
            "src/batgrl/_fbuf.pyx",
            "src/batgrl/_rendering.pyx",
            "src/batgrl/_sixel.pyx",
            "src/batgrl/char_width.pyx",
            "src/batgrl/geometry/regions.pyx",
        ]
    ),
    include_dirs=[np.get_include()],
    cmdclass={"build_py": build_py_with_cwidth},
)
